# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

from scripts.psd import IOCQ, IOSQ, PRP, PRPList, SQE, CQE


def test_pcie_identifiers(pcie):
    logging.info("Identifiers: 0x%x" % pcie.register(0))
    logging.info("Subsystem Identifiers: 0x%x" % pcie.register(0x2c))
    
    
def test_pcie_command(pcie):
    logging.info("Command: 0x%x" % pcie.register(4, 2))
    #check Memory Space Enable = 1
    assert pcie.register(4, 2)&0x2 != 0   
    #check reserved field
    assert pcie.register(4, 2)&0xf800 == 0

    
def test_pcie_revision_id(pcie):
    logging.info("Revision ID: 0x%x" % pcie.register(8, 1))

    
def test_pcie_class_code(pcie):
    logging.info("Class Code: 0x%x" % pcie.register(9, 3))
    assert pcie.register(9, 3) == 0x010802

    
def test_pcie_bist(pcie):
    logging.info("Builtin self test: 0x%x" % pcie.register(0xf, 1))
    assert pcie.register(0xf, 1) == 0
    
    
def test_pcie_pmcr(pcie):
    pmcr_addr = pcie.cap_offset(0x01)
    pmcr = pcie.register(pmcr_addr, 4)
    logging.info("pmcr register [0x%x]= 0x%x"% (pmcr_addr, pmcr))
    logging.info("Version: %d" % ((pmcr>>16)&7))
    logging.info("AUX currect %d" % ((pmcr>>22)&7))

    
def test_pcie_pmcsr(pcie):
    pmcsr_addr = pcie.cap_offset(0x01)+4
    pmcsr = pcie.register(pmcsr_addr, 4)
    logging.info("pmcsr register [0x%x]= 0x%x"% (pmcsr_addr, pmcsr))
    logging.info("power state %d" % ((pmcsr>>0)&3))
    logging.info("no soft reset: %d" % ((pmcsr>>3)&1))

    scale = (pmcsr>>13)&0x3
    if scale:
        data = (pmcsr>>24)&0xff
        logging.info("D0 power consumption: %d mW" % (data*1000*(0.1**scale)))
    

def test_pcie_pcie_cap(pcie):
    pciecap_addr = pcie.cap_offset(0x10)
    pciecap = pcie.register(pciecap_addr, 4)
    logging.info("pcie capability register [0x%x]= 0x%x"% (pciecap_addr, pciecap))
    logging.info("capability version: %d" % ((pciecap>>16)&0x7))
    logging.info("device type: %d" % ((pciecap>>20)&0xf))
    logging.info("slot: %d" % ((pciecap>>24)&0x1))

    devcap = pcie.register(pciecap_addr+4, 4)
    logging.info("device capability register: 0x%x" % devcap)
    logging.info("Max_Payload_Size Supported: %d Byte" % (128*(2**(devcap&0x7))))

    devctrl = pcie.register(pciecap_addr+8, 2)
    logging.info("device control register: 0x%x" % devctrl)
    logging.info("relaxed ordering: %d" % ((devctrl>>4)&1))
    logging.info("Max_Payload_Size: %d Byte" % (128*(2**((devctrl>>5)&0x7))))
    logging.info("Max_Read_Request_Size: %d Byte" % (128*(2**((devctrl>>12)&0x7))))

    devsts = pcie.register(pciecap_addr+10, 2)
    logging.info("device status register: 0x%x" % devsts)
    pcie[pciecap_addr+10] = 1
    devsts = pcie.register(pciecap_addr+10, 2)
    logging.info("cleared correectable error detected bit")
    logging.info("device status register: 0x%x" % devsts)

    
def test_pcie_link_capabilities_and_status(pcie):
    linkcap_addr = pcie.cap_offset(0x10)+12
    linkcap = pcie.register(linkcap_addr, 4)
    logging.info("link capability register [0x%x]= 0x%x"% (linkcap_addr, linkcap))    
    logging.info("max link speed: %d"% ((linkcap>>0)&0xf))
    logging.info("max link width: %d"% ((linkcap>>4)&0x3f))
    logging.info("ASPM Support: %d"% ((linkcap>>10)&0x3))

    linkctrl_addr = pcie.cap_offset(0x10)+16
    linkctrl = pcie.register(linkctrl_addr, 2)
    logging.info("link control register [0x%x]= 0x%x"% (linkctrl_addr, linkctrl))    
    
    linksts_addr = pcie.cap_offset(0x10)+18
    linksts = pcie.register(linkcap_addr, 2)
    logging.info("link status register [0x%x]= 0x%x"% (linksts_addr, linksts))
    logging.info("link speed: %d"% ((linksts>>0)&0xf))
    logging.info("link width: %d"% ((linksts>>4)&0x3f))
    logging.info("link training: %d"% ((linksts>>11)&0x1))
    logging.info("link active: %d"% ((linksts>>13)&0x1))

    
@pytest.mark.parametrize("aspm", [0, 2])
def test_pcie_link_control_aspm(nvme0, pcie, aspm): #1:0
    linkctrl_addr = pcie.cap_offset(0x10)+16
    linkctrl = pcie.register(linkctrl_addr, 2)
    logging.info("link control register [0x%x]= 0x%x" %
                 (linkctrl_addr, linkctrl))

    # set ASPM control
    pcie[linkctrl_addr] = (linkctrl&0xfc)|aspm
    linkctrl = pcie.register(linkctrl_addr, 2)
    logging.info("link control register [0x%x]= 0x%x" %
                 (linkctrl_addr, linkctrl))

    # IO queue for read commands
    cq = IOCQ(nvme0, 1, 16, PRP())
    sq = IOSQ(nvme0, 1, 16, PRP(), cqid=1)

    # read lba 0 for 1000 times, interleaved with delays
    read_cmd = SQE(2, 1)
    read_cmd.prp1 = PRPList()
    pbit = 1
    for i in range(100):
        logging.debug(i)
        slot = i%16
        if slot == 0:
            pbit = not pbit
        next_slot = slot+1
        if next_slot == 16:
            next_slot = 0
        sq[slot] = read_cmd
        sq.tail = next_slot
        while cq[slot].p == pbit: pass
        cq.head = next_slot

        # delay to trigger ASPM
        time.sleep(0.01)

    sq.delete()
    cq.delete()

    time.sleep(1)
    
    #return ASPM L0
    pcie[linkctrl_addr] = (linkctrl&0xfc)|0
    

def test_pcie_cold_reset(subsystem, nvme0, buf):
    nvme0.identify(buf).waitdone()
    subsystem.power_cycle()
    nvme0.reset()
    nvme0.identify(buf).waitdone()


def test_pcie_format(nvme0n1):
    nvme0n1.format()

    
def test_pcie_read_bandwidth(nvme0n1):
    io_size = 128
    r = nvme0n1.ioworker(io_size=io_size,
                         lba_random=False,
                         read_percentage=100,
                         time=1).start().close()
    logging.debug(r)
    logging.info("%dMB/s" % ((io_size*512*r.io_count_read/1000)/1000))
    

@pytest.mark.parametrize("aspm", [0, 2])
@pytest.mark.parametrize("gen", [1, 2, 3, 2, 1, 1, 2, 3])
def test_pcie_link_speed(pcie, nvme0, nvme0n1, gen, aspm):
    linkctr2_addr = pcie.cap_offset(0x10)+0x30
    linkctr2 = pcie.register(linkctr2_addr, 4)
    logging.info(linkctr2)

    pcie[linkctr2_addr] = (linkctr2 & 0xf0) | gen
    logging.info(pcie.register(linkctr2_addr, 4))
    pcie.reset()
    nvme0.reset()
    test_pcie_read_bandwidth(nvme0n1)
    test_pcie_link_control_aspm(nvme0, pcie, aspm)
    

@pytest.mark.parametrize("gen", [5, 4, 0, 3])
def _test_pcie_link_speed_invalid(pcie, nvme0, nvme0n1, gen):
    linkctr2_addr = pcie.cap_offset(0x10)+0x30
    linkctr2 = pcie.register(linkctr2_addr, 4)
    logging.info(linkctr2)

    pcie[linkctr2_addr] = (linkctr2 & 0xf0) | gen
    logging.info(pcie.register(linkctr2_addr, 4))
    pcie.reset()
    nvme0.reset()
    



#
#  BSD LICENSE
#
#  Copyright (c) Crane Chu <cranechu@gmail.com>
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in
#      the documentation and/or other materials provided with the
#      distribution.
#    * Neither the name of Intel Corporation nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# -*- coding: utf-8 -*-


import enum
import time
import pytest
import struct
import logging
import warnings

from nvme import *


class OPAL_TOKEN(enum.IntEnum):
    TRUE = 0x01
    FALSE = 0x00
    TABLE = 0x00
    STARTROW = 0x01
    ENDROW = 0x02
    STARTCOLUMN = 0x03
    ENDCOLUMN = 0x04
    VALUES = 0x01
    PIN = 0x03
    RANGESTART = 0x03
    RANGELENGTH = 0x04
    READLOCKENABLED = 0x05
    WRITELOCKENABLED = 0x06
    READLOCKED = 0x07
    WRITELOCKED = 0x08
    ACTIVEKEY = 0x0A
    MAXRANGES = 0x04
    MBRENABLE = 0x01
    MBRDONE = 0x02
    HOSTPROPERTIES = 0x00
    STARTLIST = 0xF0
    ENDLIST = 0xF1
    STARTNAME = 0xF2
    ENDNAME = 0xF3
    CALL = 0xF8
    ENDOFDATA = 0xF9
    ENDOFSESSION = 0xFA
    STARTTRANSACTON = 0xFB
    ENDTRANSACTON = 0xFC
    EMPTYATOM = 0xFF
    WHERE = 0x00
    LIFECYCLE = 0x06
    AUTH_ENABLE = 0x05
    BOOLEAN_EXPR = 0x03


class OPAL_UID(enum.IntEnum):
    # user
    SMUID = 0
    THISSP = 1
    ADMINSP = 2
    LOCKINGSP = 3
    ANYBODY = 4
    SID = 5
    ADMIN1 = 6
    USER1 = 7
    USER2 = 8

    # table
    LOCKINGRANGE_GLOBAL = 9
    LOCKINGRANGE_ACE_RDLOCKED = 10
    LOCKINGRANGE_ACE_WRLOCKED = 11
    MBRCONTROL = 12
    MBR = 13
    AUTHORITY_TABLE = 14
    C_PIN_TABLE = 15
    LOCKING_INFO_TABLE = 16
    PSID = 17

    # C_PIN
    C_PIN_MSID = 18
    C_PIN_SID = 19
    C_PIN_ADMIN1 = 20
    C_PIN_USER1 = 21

    # half
    HALF_AUTHORITY_OBJ_REF = 22
    HALF_BOOLEAN_ACE = 23


class OPAL_METHOD(enum.IntEnum):
    PROPERTIES = 0
    STARTSESSION = 1
    REVERT = 2
    ACTIVATE = 3
    NEXT = 4
    GETACL = 5
    GENKEY = 6
    REVERTSP = 7
    GET = 8
    SET = 9
    AUTHENTICATE = 10
    RANDOM = 11


opal_uid_table = {
    OPAL_UID.SMUID: [0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff],
    OPAL_UID.THISSP: [0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01],
    OPAL_UID.ADMINSP: [0x00, 0x00, 0x02, 0x05, 0x00, 0x00, 0x00, 0x01],
    OPAL_UID.LOCKINGSP: [0x00, 0x00, 0x02, 0x05, 0x00, 0x00, 0x00, 0x02],
    OPAL_UID.ANYBODY: [0x00, 0x00, 0x00, 0x09, 0x00, 0x00, 0x00, 0x01],
    OPAL_UID.SID: [0x00, 0x00, 0x00, 0x09, 0x00, 0x00, 0x00, 0x06],
    OPAL_UID.ADMIN1: [0x00, 0x00, 0x00, 0x09, 0x00, 0x01, 0x00, 0x01],
    OPAL_UID.USER1: [0x00, 0x00, 0x00, 0x09, 0x00, 0x03, 0x00, 0x01],
    OPAL_UID.USER2: [0x00, 0x00, 0x00, 0x09, 0x00, 0x03, 0x00, 0x02],
    OPAL_UID.LOCKINGRANGE_GLOBAL: [0x00, 0x00, 0x08, 0x02, 0x00, 0x00, 0x00, 0x01],
    OPAL_UID.LOCKINGRANGE_ACE_RDLOCKED: [0x00, 0x00, 0x00, 0x08, 0x00, 0x03, 0xE0, 0x01],
    OPAL_UID.LOCKINGRANGE_ACE_WRLOCKED: [0x00, 0x00, 0x00, 0x08, 0x00, 0x03, 0xE8, 0x01],
    OPAL_UID.MBRCONTROL: [0x00, 0x00, 0x08, 0x03, 0x00, 0x00, 0x00, 0x01],
    OPAL_UID.MBR: [0x00, 0x00, 0x08, 0x04, 0x00, 0x00, 0x00, 0x00],
    OPAL_UID.AUTHORITY_TABLE: [0x00, 0x00, 0x00, 0x09, 0x00, 0x00, 0x00, 0x00],
    OPAL_UID.C_PIN_TABLE: [0x00, 0x00, 0x00, 0x0B, 0x00, 0x00, 0x00, 0x00],
    OPAL_UID.LOCKING_INFO_TABLE: [0x00, 0x00, 0x08, 0x01, 0x00, 0x00, 0x00, 0x01],
    OPAL_UID.PSID: [0x00, 0x00, 0x00, 0x09, 0x00, 0x01, 0xff, 0x01],
    OPAL_UID.C_PIN_MSID: [0x00, 0x00, 0x00, 0x0B, 0x00, 0x00, 0x84, 0x02],
    OPAL_UID.C_PIN_SID: [0x00, 0x00, 0x00, 0x0B, 0x00, 0x00, 0x00, 0x01],
    OPAL_UID.C_PIN_ADMIN1: [0x00, 0x00, 0x00, 0x0B, 0x00, 0x01, 0x00, 0x01],
    OPAL_UID.C_PIN_USER1: [0x00, 0x00, 0x00, 0x0B, 0x00, 0x03, 0x00, 0x01],
    OPAL_UID.HALF_AUTHORITY_OBJ_REF: [0x00, 0x00, 0x0C, 0x05, 0xff, 0xff, 0xff, 0xff],
    OPAL_UID.HALF_BOOLEAN_ACE: [0x00, 0x00, 0x04, 0x0E, 0xff, 0xff, 0xff, 0xff]
}


opal_method_table = {
    OPAL_METHOD.PROPERTIES: [0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0x01],
    OPAL_METHOD.STARTSESSION: [0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0x02],
    OPAL_METHOD.REVERT: [0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x02, 0x02],
    OPAL_METHOD.ACTIVATE: [0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x02, 0x03],
    OPAL_METHOD.NEXT: [0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x00, 0x08],
    OPAL_METHOD.GETACL: [0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x00, 0x0d],
    OPAL_METHOD.GENKEY: [0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x00, 0x10],
    OPAL_METHOD.REVERTSP: [0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x00, 0x11],
    OPAL_METHOD.GET: [0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x00, 0x16],
    OPAL_METHOD.SET: [0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x00, 0x17],
    OPAL_METHOD.AUTHENTICATE: [0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x00, 0x1c],
    OPAL_METHOD.RANDOM: [0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x06, 0x01]
}


class Command(object):
    def __init__(self, nvme0, comid):
        self.buf = Buffer(2048, "TCG Command Buffer")
        self.nvme0 = nvme0
        self.comid = comid
        self.buf[:8] = struct.pack('>IHH', 0, comid, 0)
        self.pos = 0x38  # append position

    def send(self, append_end_tokens=True):
        if append_end_tokens:
            self.append_token_list(OPAL_TOKEN.ENDOFDATA,
                                   OPAL_TOKEN.STARTLIST,
                                   0, 0, 0,
                                   OPAL_TOKEN.ENDLIST)

        # fill length
        assert self.pos > 56  # larger than header
        self.buf[52:] = struct.pack('>I', self.pos-56)
        while self.pos % 4:
            # padding to dword
            self.append_u8(0)
        self.buf[16:] = struct.pack('>I', self.pos-20)
        self.buf[40:] = struct.pack('>I', self.pos-44)
        assert self.pos < 2048

        # send packet
        logging.debug(self.buf.dump(256))
        self.nvme0.security_send(self.buf, self.comid).waitdone()
        return self

    def append_u8(self, val):
        self.buf[self.pos] = val
        self.pos += 1

    def append_u16(self, val):
        self.buf[self.pos:] = struct.pack('>H', val)
        self.pos += 2

    def append_u32(self, val):
        self.buf[self.pos:] = struct.pack('>I', val)
        self.pos += 4

    def append_u64(self, val):
        self.buf[self.pos:] = struct.pack('>Q', val)
        self.pos += 8

    def append_token(self, token):
        self.append_u8(token)

    def append_token_uid(self, u):
        uid = opal_uid_table[u]
        self.append_u8(0xa0+len(uid))
        self.append_token_list(*uid)

    def append_token_method(self, m):
        method = opal_method_table[m]
        self.append_u8(0xa0+len(method))
        self.append_token_list(*method)

    def append_token_list(self, *val_list):
        for val in val_list:
            self.append_u8(int(val))

    def append_token_atom(self, atom):
        if type(atom) == int:
            if atom < 64:
                # tiny
                self.append_u8(atom)
            elif atom < 0x100:
                # short: 1-byte int
                self.append_u8(0x81)
                self.append_u8(atom)
            elif atom < 0x10000:
                # short: 2-byte int
                self.append_u8(0x82)
                self.append_u16(atom)
            elif atom < 0x100000000:
                # medium: 4-byte int
                self.append_u8(0x84)
                self.append_u32(atom)
            else:
                # long: 8-byte int
                self.append_u8(0x88)
                self.append_u64(atom)
        else:
            assert type(atom) == bytes
            if len(atom) < 16:
                # short
                self.append_u8(0xa0 + len(atom))
            elif len(atom):
                # medium
                self.append_u8(0xd0)
                self.append_u8(len(atom))
            self.append_token_list(*list(atom))

    def _start_session(self, hsn):
        self.append_token(OPAL_TOKEN.CALL)
        self.append_token_uid(OPAL_UID.SMUID)
        self.append_token_method(OPAL_METHOD.STARTSESSION)
        self.append_token(OPAL_TOKEN.STARTLIST)
        self.append_token_atom(hsn)

    def start_anybody_adminsp_session(self, hsn):
        self._start_session(hsn)
        self.append_token_uid(OPAL_UID.ADMINSP)
        self.append_token(OPAL_TOKEN.TRUE)
        self.append_token(OPAL_TOKEN.ENDLIST)
        return self

    def start_adminsp_session(self, hsn, key):
        self._start_session(hsn)
        self.append_token_uid(OPAL_UID.ADMINSP)
        self.append_token(OPAL_TOKEN.TRUE)
        self.append_token(OPAL_TOKEN.STARTNAME)
        self.append_token(0)
        self.append_token_atom(key)
        self.append_token_list(OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.STARTNAME,
                               3)
        self.append_token_uid(OPAL_UID.SID)
        self.append_token(OPAL_TOKEN.ENDNAME)
        self.append_token(OPAL_TOKEN.ENDLIST)
        return self

    def start_auth_session(self, hsn, user_id, key):
        self._start_session(hsn)
        self.append_token_uid(OPAL_UID.LOCKINGSP)
        self.append_token(OPAL_TOKEN.TRUE)
        self.append_token(OPAL_TOKEN.STARTNAME)
        self.append_token(0)
        self.append_token_atom(key)
        self.append_token_list(OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.STARTNAME,
                               3)
        if user_id == 0:
            self.append_token_uid(OPAL_UID.ADMIN1)
        elif user_id == 1:
            self.append_token_uid(OPAL_UID.USER1)
        elif user_id == 2:
            self.append_token_uid(OPAL_UID.USER2)
        else:
            assert False, "not supported user id %d" % user_id
        self.append_token(OPAL_TOKEN.ENDNAME)
        self.append_token(OPAL_TOKEN.ENDLIST)
        return self

    def revert_tper(self, hsn, tsn):
        self.append_token(OPAL_TOKEN.CALL)
        self.append_token_uid(OPAL_UID.ADMINSP)
        self.append_token_method(OPAL_METHOD.REVERT)
        self.append_token(OPAL_TOKEN.STARTLIST)
        self.append_token(OPAL_TOKEN.ENDLIST)
        self.buf[20:] = struct.pack('>I', tsn)
        self.buf[24:] = struct.pack('>I', hsn)
        return self

    def get_msid_cpin_pin(self, hsn, tsn):
        self.append_token(OPAL_TOKEN.CALL)
        self.append_token_uid(OPAL_UID.C_PIN_MSID)
        self.append_token_method(OPAL_METHOD.GET)
        self.append_token_list(OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.STARTCOLUMN,
                               OPAL_TOKEN.PIN,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.ENDCOLUMN,
                               OPAL_TOKEN.PIN,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.ENDLIST,
                               OPAL_TOKEN.ENDLIST)
        self.buf[20:] = struct.pack('>I', tsn)
        self.buf[24:] = struct.pack('>I', hsn)
        return self

    def get_locking_sp_lifecycle(self, hsn, tsn):
        self.append_token(OPAL_TOKEN.CALL)
        self.append_token_uid(OPAL_UID.LOCKINGSP)
        self.append_token_method(OPAL_METHOD.GET)
        self.append_token_list(OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.STARTCOLUMN,
                               OPAL_TOKEN.LIFECYCLE,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.ENDCOLUMN,
                               OPAL_TOKEN.LIFECYCLE,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.ENDLIST,
                               OPAL_TOKEN.ENDLIST)
        self.buf[20:] = struct.pack('>I', tsn)
        self.buf[24:] = struct.pack('>I', hsn)
        return self

    def activate(self, hsn, tsn):
        self.append_token(OPAL_TOKEN.CALL)
        self.append_token_uid(OPAL_UID.LOCKINGSP)
        self.append_token_method(OPAL_METHOD.ACTIVATE)
        self.append_token_list(OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.ENDLIST)
        self.buf[20:] = struct.pack('>I', tsn)
        self.buf[24:] = struct.pack('>I', hsn)
        return self

    def setup_range(self, hsn, tsn, range, slba, nlb):
        range_uid = opal_uid_table[OPAL_UID.LOCKINGRANGE_GLOBAL][:]
        if range:  # not global
            range_uid[5] = 3
            range_uid[7] = range

        self.append_token(OPAL_TOKEN.CALL)
        self.append_u8(0xa0+len(range_uid))
        self.append_token_list(*range_uid)
        self.append_token_method(OPAL_METHOD.SET)
        self.append_token_list(OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.VALUES,
                               OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.RANGESTART)
        self.append_token_atom(slba)
        self.append_token_list(OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.RANGELENGTH)
        self.append_token_atom(nlb)
        self.append_token_list(OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.READLOCKENABLED,
                               OPAL_TOKEN.TRUE,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.WRITELOCKENABLED,
                               OPAL_TOKEN.TRUE,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.ENDLIST,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.ENDLIST)
        self.buf[20:] = struct.pack('>I', tsn)
        self.buf[24:] = struct.pack('>I', hsn)
        return self

    def gen_new_key(self, hsn, tsn, range, prev_data):
        self.append_token(OPAL_TOKEN.CALL)
        self.append_u8(0xa0+len(prev_data))
        self.append_token_list(*prev_data)
        self.append_token_method(OPAL_METHOD.GENKEY)
        self.append_token_list(OPAL_TOKEN.STARTLIST, OPAL_TOKEN.ENDLIST)
        self.buf[20:] = struct.pack('>I', tsn)
        self.buf[24:] = struct.pack('>I', hsn)
        return self

    def get_active_key(self, hsn, tsn, range):
        range_uid = opal_uid_table[OPAL_UID.LOCKINGRANGE_GLOBAL][:]
        if range:  # not global
            range_uid[5] = 3
            range_uid[7] = range

        self.append_token(OPAL_TOKEN.CALL)
        self.append_u8(0xa0+len(range_uid))
        self.append_token_list(*range_uid)
        self.append_token_method(OPAL_METHOD.GET)
        self.append_token_list(OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.STARTCOLUMN,
                               OPAL_TOKEN.ACTIVEKEY,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.ENDCOLUMN,
                               OPAL_TOKEN.ACTIVEKEY,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.ENDLIST,
                               OPAL_TOKEN.ENDLIST)
        self.buf[20:] = struct.pack('>I', tsn)
        self.buf[24:] = struct.pack('>I', hsn)
        return self

    def lock_unlock_range(self, hsn, tsn, range, rlock, wlock):
        range_uid = opal_uid_table[OPAL_UID.LOCKINGRANGE_GLOBAL][:]
        if range:  # not global
            range_uid[5] = 3
            range_uid[7] = range

        self.append_token(OPAL_TOKEN.CALL)
        self.append_u8(0xa0+len(range_uid))
        self.append_token_list(*range_uid)
        self.append_token_method(OPAL_METHOD.SET)
        self.append_token_list(OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.VALUES,
                               OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.READLOCKED,
                               rlock,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.WRITELOCKED,
                               wlock,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.ENDLIST,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.ENDLIST)
        self.buf[20:] = struct.pack('>I', tsn)
        self.buf[24:] = struct.pack('>I', hsn)
        return self

    def add_user_to_range(self, hsn, tsn, user, range, passwd, can_write=True):
        user_uid = opal_uid_table[OPAL_UID.USER1][:]
        user_uid[7] = user
        range_uid = opal_uid_table[OPAL_UID.LOCKINGRANGE_ACE_RDLOCKED+can_write][:]
        range_uid[7] = range
        auth_ref_uid = opal_uid_table[OPAL_UID.HALF_AUTHORITY_OBJ_REF][:4]
        boolean_ace_uid = opal_uid_table[OPAL_UID.HALF_BOOLEAN_ACE][:4]

        self.append_token(OPAL_TOKEN.CALL)
        self.append_u8(0xa0+len(range_uid))
        self.append_token_list(*range_uid)
        self.append_token_method(OPAL_METHOD.SET)
        self.append_token_list(OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.VALUES,
                               OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.BOOLEAN_EXPR,
                               OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTNAME)

        self.append_u8(0xa0+len(auth_ref_uid))
        self.append_token_list(*auth_ref_uid)
        self.append_u8(0xa0+len(user_uid))
        self.append_token_list(*user_uid)
        self.append_token_list(OPAL_TOKEN.ENDNAME, OPAL_TOKEN.STARTNAME)
        self.append_u8(0xa0+len(auth_ref_uid))
        self.append_token_list(*auth_ref_uid)
        self.append_u8(0xa0+len(user_uid))
        self.append_token_list(*user_uid)
        self.append_token_list(OPAL_TOKEN.ENDNAME, OPAL_TOKEN.STARTNAME)
        self.append_u8(0xa0+len(boolean_ace_uid))
        self.append_token_list(*boolean_ace_uid)
        self.append_token_list(OPAL_TOKEN.TRUE,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.ENDLIST,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.ENDLIST,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.ENDLIST)
        self.buf[20:] = struct.pack('>I', tsn)
        self.buf[24:] = struct.pack('>I', hsn)
        return self

    def enable_user(self, hsn, tsn, user_id):
        user_uid = opal_uid_table[OPAL_UID.USER1][:]
        user_uid[7] = user_id

        self.append_token(OPAL_TOKEN.CALL)
        self.append_u8(0xa0+len(user_uid))
        self.append_token_list(*user_uid)
        self.append_token_method(OPAL_METHOD.SET)
        self.append_token_list(OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.VALUES,
                               OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.AUTH_ENABLE,
                               OPAL_TOKEN.TRUE,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.ENDLIST,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.ENDLIST)
        self.buf[20:] = struct.pack('>I', tsn)
        self.buf[24:] = struct.pack('>I', hsn)
        return self

    def set_sid_cpin_pin(self, hsn, tsn, new_passwd):
        self.append_token(OPAL_TOKEN.CALL)
        self.append_token_uid(OPAL_UID.C_PIN_SID)
        self.append_token_method(OPAL_METHOD.SET)
        self.append_token_list(OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.VALUES,
                               OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.PIN)
        self.append_token_atom(new_passwd)
        self.append_token_list(OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.ENDLIST,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.ENDLIST)
        self.buf[20:] = struct.pack('>I', tsn)
        self.buf[24:] = struct.pack('>I', hsn)
        return self

    def set_new_passwd(self, hsn, tsn, user_id, new_passwd):
        self.append_token(OPAL_TOKEN.CALL)
        if user_id == 0:
            self.append_token_uid(OPAL_UID.C_PIN_ADMIN1)
        elif user_id == 1:
            self.append_token_uid(OPAL_UID.C_PIN_USER1)
        else:
            assert False, "not supported user id %d" % user_id
        self.append_token_method(OPAL_METHOD.SET)
        self.append_token_list(OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.VALUES,
                               OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTNAME,
                               OPAL_TOKEN.PIN)
        self.append_token_atom(new_passwd)
        self.append_token_list(OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.ENDLIST,
                               OPAL_TOKEN.ENDNAME,
                               OPAL_TOKEN.ENDLIST)
        self.buf[20:] = struct.pack('>I', tsn)
        self.buf[24:] = struct.pack('>I', hsn)
        return self

    def end_session(self, hsn, tsn):
        self.append_u8(OPAL_TOKEN.ENDOFSESSION)
        self.buf[20:] = struct.pack('>I', tsn)
        self.buf[24:] = struct.pack('>I', hsn)
        return self

    def properties(self, host_properties={}):
        self.append_token(OPAL_TOKEN.CALL)
        self.append_token_uid(OPAL_UID.SMUID)
        self.append_token_method(OPAL_METHOD.PROPERTIES)
        self.append_token_list(OPAL_TOKEN.STARTLIST,
                               OPAL_TOKEN.STARTNAME,
                               0)  # host properties list
        self.append_token(OPAL_TOKEN.STARTLIST)

        for k in host_properties:
            self.append_token(OPAL_TOKEN.STARTNAME)
            self.append_token_atom(k)
            self.append_token_atom(host_properties[k])
            self.append_token(OPAL_TOKEN.ENDNAME)

        self.append_token(OPAL_TOKEN.ENDLIST)
        self.append_token(OPAL_TOKEN.ENDNAME)
        self.append_token(OPAL_TOKEN.ENDLIST)

        return self


class Response(object):
    def __init__(self, nvme0, comid=1):
        self.nvme0 = nvme0
        self.comid = comid
        self.buf = Buffer(2048, "TCG Response Buffer")
        self.parsed = []

    def receive(self, check_status=True):
        self.nvme0.security_receive(self.buf, self.comid).waitdone()
        logging.debug(self.buf.dump(256))

        def _parse_token(ptr, len, bytes):
            if bytes:
                # byte sequence
                self.parsed.append(self.buf[ptr+1:ptr+1+len])
            else:
                # int
                if len == 2:
                    pattern = ">H"
                elif len == 4:
                    pattern = ">I"
                else:
                    logging.info(len)
                    assert False
                l = struct.unpack(pattern, self.buf[ptr+1:ptr+1+len])
                self.parsed.append(l[0])

        # parse the buffer
        ptr = 0x38
        _size = struct.unpack(">H", self.buf[0x36:0x38])[0]
        while ptr < 0x38+_size:
            ch = self.buf.data(ptr)

            if ch < 0x80:
                # tiny
                self.parsed.append(ch)
            elif ch >= 0x80 and ch < 0xc0:
                # short atom
                _len = ch & 0xf
                _parse_token(ptr, _len, ch & 0x20)
                ptr += _len
            elif ch >= 0xc0 and ch < 0xe0:
                # medium
                _len = ((ch & 7) << 8)+self.buf.data(ptr+1)
                ptr += 1
                _parse_token(ptr, _len, ch & 0x10)
                ptr += _len
            elif ch >= 0xe0 and ch < 0xe4:
                # medium
                _len = self.buf.data(ptr+1, ptr+4)
                ptr += 3
                _parse_token(ptr, _len, ch & 0x2)
                ptr += _len

            ptr += 1

        # check status
        logging.debug(self.parsed)
        if check_status and len(self.parsed) >= 3:
            if self.parsed[-3]:
                warnings.warn("TCG response error: "+str(self.parsed))
        return self

    def level0_discovery(self):
        comid = 0
        total_length, ver, _ = struct.unpack('>IIQ', self.buf[:16])
        total_length += 4
        offset = 48
        while offset < total_length:
            feature, version, length = struct.unpack(
                '>HBB', self.buf[offset:offset+4])
            version >>= 4
            length += 4

            # parse discovery response buffer
            if feature == 0x0303 or \
               feature == 0x0302:
                # pyrite 2, or pyrite 1
                comid = struct.unpack('>H', self.buf[offset+4:offset+6])[0]

            offset += length
        assert offset == total_length
        assert comid, "no pyrite feature found"
        return comid

    def start_session(self):
        return self.parsed[2], self.parsed[3]

    def get_c_pin_msid(self):
        length = struct.unpack(">B", self.buf[0x3c:0x3d])[0]
        length = length & 0xf
        return self.parsed[1]

    def get_active_key(self):
        return self.parsed[1]


def test_properties(nvme0):
    host_properties = {
        b"MaxComPacketSize": 4096,
        b"MaxPacketSize": 4076,
        b"MaxIndTokenSize": 4040,
        b"MaxPackets": 1,
        b"MaxSubPackets": 1,
        b"MaxMethods": 1,
    }
    comid = Response(nvme0).receive(False).level0_discovery()
    Command(nvme0, comid).properties().send()
    Response(nvme0, comid).receive(False)
    Command(nvme0, comid).properties(host_properties).send()
    Response(nvme0, comid).receive(False)


def test_tcg_all_basic_function(subsystem, nvme0, nvme0n1, qpair, buf, new_passwd=b'123456'):
    # subsystem.power_cycle()
    # nvme0.reset()

    nvme0n1.write(qpair, buf, 0).waitdone()
    nvme0n1.write(qpair, buf, 128).waitdone()
    nvme0n1.read(qpair, buf, 0).waitdone()
    nvme0n1.read(qpair, buf, 128).waitdone()

    comid = Response(nvme0).receive().level0_discovery()

    logging.info("test: take ownership")
    Command(nvme0, comid).start_anybody_adminsp_session(0x69).send()
    hsn, tsn = Response(nvme0, comid).receive().start_session()
    logging.debug("hsn 0x%x, tsn 0x%x" % (hsn, tsn))
    Command(nvme0, comid).get_msid_cpin_pin(hsn, tsn).send()
    password = Response(nvme0, comid).receive().get_c_pin_msid()
    Command(nvme0, comid).end_session(hsn, tsn).send(False)
    Response(nvme0, comid).receive()
    nvme0n1.read(qpair, buf, 0).waitdone()

    logging.info("test: set password")
    Command(nvme0, comid).start_adminsp_session(0x69, password).send()
    hsn, tsn = Response(nvme0, comid).receive().start_session()
    Command(nvme0, comid).set_sid_cpin_pin(hsn, tsn, new_passwd).send()
    Response(nvme0, comid).receive()
    Command(nvme0, comid).end_session(hsn, tsn).send(False)
    Response(nvme0, comid).receive()

    logging.info("test: activate locking sp")
    Command(nvme0, comid).start_adminsp_session(0x69, new_passwd).send()
    hsn, tsn = Response(nvme0, comid).receive().start_session()
    Command(nvme0, comid).get_locking_sp_lifecycle(hsn, tsn).send()
    Response(nvme0, comid).receive()
    Command(nvme0, comid).activate(hsn, tsn).send()
    Response(nvme0, comid).receive()
    Command(nvme0, comid).end_session(hsn, tsn).send(False)
    Response(nvme0, comid).receive()
    nvme0n1.read(qpair, buf, 0).waitdone()

    logging.info("test: setup range")
    Command(nvme0, comid).start_auth_session(0x69, 0, new_passwd).send()
    hsn, tsn = Response(nvme0, comid).receive().start_session()
    Command(nvme0, comid).setup_range(hsn, tsn, 1, 0, 128).send()
    Response(nvme0, comid).receive()
    Command(nvme0, comid).end_session(hsn, tsn).send(False)
    Response(nvme0, comid).receive()
    nvme0n1.read(qpair, buf, 0).waitdone()

    logging.info("test: enable user")
    Command(nvme0, comid).start_auth_session(0x69, 0, new_passwd).send()
    hsn, tsn = Response(nvme0, comid).receive().start_session()
    Command(nvme0, comid).enable_user(hsn, tsn, 1).send()
    Response(nvme0, comid).receive()
    Command(nvme0, comid).end_session(hsn, tsn).send(False)
    Response(nvme0, comid).receive()
    nvme0n1.read(qpair, buf, 0).waitdone()

    logging.info("test: change passwd")
    Command(nvme0, comid).start_auth_session(0x69, 0, new_passwd).send()
    hsn, tsn = Response(nvme0, comid).receive().start_session()
    Command(nvme0, comid).set_new_passwd(hsn, tsn, 1, b"654321").send()
    Response(nvme0, comid).receive()
    Command(nvme0, comid).end_session(hsn, tsn).send(False)
    Response(nvme0, comid).receive()

    logging.info("test: user session")
    Command(nvme0, comid).start_auth_session(0x69, 1, b"654321").send()
    hsn, tsn = Response(nvme0, comid).receive().start_session()
    Command(nvme0, comid).set_new_passwd(hsn, tsn, 1, b"111111").send()
    Response(nvme0, comid).receive()
    Command(nvme0, comid).end_session(hsn, tsn).send(False)
    Response(nvme0, comid).receive()
    nvme0n1.read(qpair, buf, 0).waitdone()

    logging.info("test: add user to range, readonly")
    Command(nvme0, comid).start_auth_session(0x69, 0, new_passwd).send()
    hsn, tsn = Response(nvme0, comid).receive().start_session()
    Command(nvme0, comid).add_user_to_range(hsn, tsn, 1, 1, new_passwd, False).send()
    Response(nvme0, comid).receive()
    Command(nvme0, comid).end_session(hsn, tsn).send(False)
    Response(nvme0, comid).receive()
    nvme0n1.read(qpair, buf, 0).waitdone()

    logging.info("test: add user to range, readwrite")
    Command(nvme0, comid).start_auth_session(0x69, 0, new_passwd).send()
    hsn, tsn = Response(nvme0, comid).receive().start_session()
    Command(nvme0, comid).add_user_to_range(hsn, tsn, 1, 1, new_passwd, True).send()
    Response(nvme0, comid).receive()
    Command(nvme0, comid).end_session(hsn, tsn).send(False)
    Response(nvme0, comid).receive()
    nvme0n1.read(qpair, buf, 0).waitdone()

    logging.info("test: unlock range, none")
    Command(nvme0, comid).start_auth_session(0x69, 1, b"111111").send()
    hsn, tsn = Response(nvme0, comid).receive().start_session()
    Command(nvme0, comid).lock_unlock_range(hsn, tsn, 1, True, True).send()
    Response(nvme0, comid).receive()
    Command(nvme0, comid).end_session(hsn, tsn).send(False)
    Response(nvme0, comid).receive()
    with pytest.warns(UserWarning, match="ERROR status: 02/86"):
        nvme0n1.read(qpair, buf, 0).waitdone()
    nvme0n1.read(qpair, buf, 128).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 02/86"):
        nvme0n1.write(qpair, buf, 0).waitdone()
    nvme0n1.write(qpair, buf, 128).waitdone()

    logging.info("test: unlock range, write only")
    Command(nvme0, comid).start_auth_session(0x69, 1, b"111111").send()
    hsn, tsn = Response(nvme0, comid).receive().start_session()
    Command(nvme0, comid).lock_unlock_range(hsn, tsn, 1, True, False).send()
    Response(nvme0, comid).receive()
    Command(nvme0, comid).end_session(hsn, tsn).send(False)
    Response(nvme0, comid).receive()
    with pytest.warns(UserWarning, match="ERROR status: 02/86"):
        nvme0n1.read(qpair, buf, 0).waitdone()
    nvme0n1.read(qpair, buf, 128).waitdone()
    nvme0n1.write(qpair, buf, 0).waitdone()
    nvme0n1.write(qpair, buf, 128).waitdone()

    logging.info("test: unlock range, read only")
    Command(nvme0, comid).start_auth_session(0x69, 1, b"111111").send()
    hsn, tsn = Response(nvme0, comid).receive().start_session()
    Command(nvme0, comid).lock_unlock_range(hsn, tsn, 1, False, True).send()
    Response(nvme0, comid).receive()
    Command(nvme0, comid).end_session(hsn, tsn).send(False)
    Response(nvme0, comid).receive()
    nvme0n1.read(qpair, buf, 0).waitdone()
    nvme0n1.read(qpair, buf, 128).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 02/86"):
        nvme0n1.write(qpair, buf, 0).waitdone()
    nvme0n1.write(qpair, buf, 128).waitdone()

    logging.info("test: unlock range, read/write")
    Command(nvme0, comid).start_auth_session(0x69, 1, b"111111").send()
    hsn, tsn = Response(nvme0, comid).receive().start_session()
    Command(nvme0, comid).lock_unlock_range(hsn, tsn, 1, False, False).send()
    Response(nvme0, comid).receive()
    Command(nvme0, comid).end_session(hsn, tsn).send(False)
    Response(nvme0, comid).receive()
    nvme0n1.write(qpair, buf, 0).waitdone()
    nvme0n1.write(qpair, buf, 1).waitdone()
    nvme0n1.write(qpair, buf, 128).waitdone()
    nvme0n1.read(qpair, buf, 0).waitdone()
    logging.info(buf.dump(16))
    nvme0n1.read(qpair, buf, 1).waitdone()
    logging.info(buf.dump(16))
    nvme0n1.read(qpair, buf, 128).waitdone()
    logging.info(buf.dump(16))

    logging.info("test: erase range")
    Command(nvme0, comid).start_auth_session(0x69, 0, new_passwd).send()
    hsn, tsn = Response(nvme0, comid).receive().start_session()
    Command(nvme0, comid).get_active_key(hsn, tsn, 1).send()
    prev_data = Response(nvme0, comid).receive().get_active_key()
    Command(nvme0, comid).gen_new_key(hsn, tsn, 1, prev_data).send()
    Response(nvme0, comid).receive()
    Command(nvme0, comid).end_session(hsn, tsn).send(False)
    Response(nvme0, comid).receive()
    nvme0n1.read(qpair, buf, 1).waitdone()
    logging.info(buf.dump(16))
    nvme0n1.read(qpair, buf, 128).waitdone()
    logging.info(buf.dump(16))

    logging.info("test: revert")
    orig_timeout = nvme0.timeout
    nvme0.timeout = 100000
    Command(nvme0, comid).start_adminsp_session(0x69, new_passwd).send()
    hsn, tsn = Response(nvme0, comid).receive().start_session()
    Command(nvme0, comid).revert_tper(hsn, tsn).send()
    Response(nvme0, comid).receive()
    # No "end session" for revert tper
    nvme0.timeout = orig_timeout
    nvme0n1.read(qpair, buf, 1).waitdone()
    logging.info(buf.dump(16))
    nvme0n1.read(qpair, buf, 128).waitdone()
    logging.info(buf.dump(16))



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem
from scripts.psd import IOCQ, IOSQ, PRP, PRPList, SQE, CQE


# PCIe, different of power states and resets
def test_power_and_reset(pcie, nvme0, subsystem):
    pcie.aspm = 2              # ASPM L1
    pcie.power_state = 3       # PCI PM D3hot
    pcie.aspm = 0
    pcie.power_state = 0

    nvme0.reset()              # controller reset: CC.EN
    nvme0.getfeatures(7).waitdone()

    pcie.reset()               # PCIe reset: hot reset, TS1, TS2
    nvme0.reset()              # reset controller after pcie reset
    nvme0.getfeatures(7).waitdone()

    subsystem.reset()          # NVMe subsystem reset: NSSR
    nvme0.reset()              # controller reset: CC.EN
    nvme0.getfeatures(7).waitdone()

    subsystem.power_cycle(10)  # power cycle NVMe device: cold reset
    nvme0.reset()              # controller reset: CC.EN
    nvme0.getfeatures(7).waitdone()

    subsystem.poweroff()
    subsystem.poweron()
    nvme0.reset()              # controller reset: CC.EN
    nvme0.getfeatures(7).waitdone()


@pytest.mark.parametrize("delay", [1]) #, 0.1, 0.01, 0.001, 0.0001, 0])
def test_reset_with_outstanding_io(nvme0, nvme0n1, delay, io_count=1000):
    nvme0n1.format(512)
    logging.debug("format done")
    
    cq = IOCQ(nvme0, 1, 1024, PRP(1024*64))
    sq = IOSQ(nvme0, 1, 1024, PRP(1024*64), cqid=1)

    write_cmd = SQE(1, 1)
    write_cmd[12] = 7  # 4K write
    buf_list = []
    for i in range(io_count):
        buf = PRP(ptype=32, pvalue=i)  # use cid as data pattern
        buf_list.append(buf)
        write_cmd.prp1 = buf
        write_cmd.cid = i
        write_cmd[10] = i
        sq[i] = write_cmd
    sq.tail = io_count

    # reset while io is active
    logging.debug("write done")
    time.sleep(delay)
    nvme0.reset()
    
    # read after reset with outstanding writes
    cq = IOCQ(nvme0, 1, 1024, PRP(1024*64))
    sq = IOSQ(nvme0, 1, 1024, PRP(1024*64), cqid=1)

    read_cmd = SQE(2, 1)
    read_cmd[12] = 7  # 4K read
    buf_list = []
    for i in range(io_count):
        buf = PRP()
        buf_list.append(buf)
        read_cmd.prp1 = buf
        read_cmd.cid = i
        read_cmd[10] = i
        sq[i] = read_cmd
    sq.tail = io_count
    logging.debug("read done")

    # data verify
    time.sleep(3)
    assert cq[io_count-1].p != 0
    for i in range(io_count):
        cid = cq[i].cid
        dp = buf_list[cq[i].cid].data(3, 0)  # check data pattern
        logging.debug("cpl %d: cid %d, data pattern %d" % (i, cid, dp))
        assert dp == 0 or dp == cid
        
    sq.delete()
    cq.delete()
    



# for pypi package information
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pynvme",
    version="2.1.3",
    author="Crane Chu",
    author_email="cranechu@gmail.com",
    description="builds your own tests.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pynvme/pynvme",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: C",
        "Programming Language :: Python :: 3.5",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires='>=3.5',
    install_requires=['pytest',
                      'pytest-excel',
                      'pylspci',
                      'quarchpy',
                      'pytemperature'],
    data_files=[
        ('pynvme',
         ['nvme.so',
          'Makefile',
          'conftest.py',
          'driver_test.py',
          'pytest.ini']),
        ('pynvme/src',
         ['src/common.sh',
          'src/setup.sh']),
        ('pynvme/scripts',
         ['scripts/psd.py',
          'scripts/tcg.py',
          'scripts/test_examples.py',
          'scripts/test_utilities.py']),
        ('pynvme/scripts/stress',
         ['scripts/stress/dirty_power_cycle_test.py']),
        ('pynvme/include/spdk',
         ['include/spdk/pci_ids.h']),
        ('pynvme/scripts/conformance/01_admin',
         ['scripts/conformance/01_admin/abort_test.py',
          'scripts/conformance/01_admin/identify_test.py',
          'scripts/conformance/01_admin/queue_test.py',
          'scripts/conformance/01_admin/firmware_test.py',
          'scripts/conformance/01_admin/dst_test.py',
          'scripts/conformance/01_admin/format_test.py',
          'scripts/conformance/01_admin/aer_test.py',
          'scripts/conformance/01_admin/sanitize_test.py',
          'scripts/conformance/01_admin/mi_test.py',
          'scripts/conformance/01_admin/features_test.py',
          'scripts/conformance/01_admin/logpage_test.py'
          ]),
        ('pynvme/scripts/conformance/02_nvm',
         ['scripts/conformance/02_nvm/compare_test.py',
          'scripts/conformance/02_nvm/flush_test.py',
          'scripts/conformance/02_nvm/read_test.py',
          'scripts/conformance/02_nvm/write_uncorrectable_test.py',
          'scripts/conformance/02_nvm/deallocate_test.py',
          'scripts/conformance/02_nvm/write_test.py',
          'scripts/conformance/02_nvm/verify_test.py',
          'scripts/conformance/02_nvm/write_zeroes_test.py'
          ]),
        ('pynvme/scripts/conformance/03_features',
         ['scripts/conformance/03_features/hmb_test.py',
          'scripts/conformance/03_features/write_protect_test.py',
          'scripts/conformance/03_features/power_management_test.py',
          'scripts/conformance/03_features/reset_test.py'
          ]),
        ('pynvme/scripts/conformance/04_registers',
         ['scripts/conformance/04_registers/controller_test.py',
          'scripts/conformance/04_registers/power_test.py',
          'scripts/conformance/04_registers/pcie_test.py'
          ]), 
        ('pynvme/scripts/conformance/05_controller',
         ['scripts/conformance/05_controller/sq_cq_test.py',
          'scripts/conformance/05_controller/sqe_cqe_test.py',
          'scripts/conformance/05_controller/interrupt_test.py',
          'scripts/conformance/05_controller/prp_test.py',
          'scripts/conformance/05_controller/arbitration_test.py'
          ]),
        ('pynvme/scripts/conformance/06_tcg',
         ['scripts/conformance/06_tcg/use_case_test.py'
          ]),
        ]
)



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem
from scripts.psd import IOCQ, IOSQ, PRP, PRPList, SQE, CQE


@pytest.fixture(scope="function")
def ncqa(nvme0):
    num_of_queue = 0
    def test_greater_id(cdw0, status):
        nonlocal num_of_queue
        num_of_queue = 1+(cdw0&0xffff)
    nvme0.getfeatures(7, cb=test_greater_id).waitdone()
    logging.info("number of queue: %d" % num_of_queue)
    return num_of_queue


@pytest.fixture(scope="function")
def mqes(nvme0):
    num_of_entry = (nvme0.cap&0xffff) + 1
    logging.info("number of entry: %d" % num_of_entry)
    return num_of_entry


def test_create_cq_basic_operation(nvme0, nvme0n1, buf, qpair):
    for i in range(10):
        nvme0n1.read(qpair, buf, 0)
    qpair.waitdone(10)


def test_create_cq_with_invalid_id(nvme0, ncqa):
    # pass case
    cq = IOCQ(nvme0, 5, 10, PRP(4096))

    # cqid: 0
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        IOCQ(nvme0, 0, 10, PRP(4096))

    # cqid: 0xffff
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        IOCQ(nvme0, 0xffff, 10, PRP(4096))

    # cqid: larger than supported number of queue
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        IOCQ(nvme0, ncqa+1, 10, PRP(4096))

    # cqid: duplicated cqid
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        IOCQ(nvme0, 5, 10, PRP(4096))
    cq.delete()

    # cqid: 0xff
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        IOCQ(nvme0, ncqa+0xff, 10, PRP(4096))


def test_create_sq_with_invalid_id(nvme0, ncqa):
    # helper cq: id 1
    cq = IOCQ(nvme0, 1, 10, PRP(4096))

    # pass case
    sq = IOSQ(nvme0, 5, 10, PRP(4096), cqid=1)

    # sqid: 0
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        IOSQ(nvme0, 0, 10, PRP(4096), cqid=1)

    # sqid: 0xffff
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        IOSQ(nvme0, 0xffff, 10, PRP(4096), cqid=1)

    # sqid: larger than supported number of queue
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        IOSQ(nvme0, ncqa+1, 10, PRP(4096), cqid=1)

    # sqid: duplicated cqid
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        IOSQ(nvme0, 5, 10, PRP(4096), cqid=1)

    # sqid: 0
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        IOSQ(nvme0, 0, 10, PRP(4096), cqid=1)

    # sqid: 0xdead
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        IOSQ(nvme0, 0xdead, 10, PRP(4096), cqid=1)

    # sqid: 0xff
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        IOSQ(nvme0, ncqa+0xff, 10, PRP(4096), cqid=1)

    sq.delete()
    cq.delete()


def test_delete_cq_with_invalid_id(nvme0, ncqa):
    def delete_cq(nvme0, qid):
        nvme0.send_cmd(0x04, cdw10 = qid).waitdone()

    # pass case
    IOCQ(nvme0, 5, 10, PRP(4096))
    delete_cq(nvme0, 5)

    # cqid: 0
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        delete_cq(nvme0, 0)

    # cqid: 0xffff
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        delete_cq(nvme0, 0xffff)

    # cqid: larger than supported number of queue
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        delete_cq(nvme0, ncqa+1)

    # cqid: duplicated cqid
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        delete_cq(nvme0, ncqa+0xff)

    # cqid: not existed
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        delete_cq(nvme0, 5)


def test_delete_sq_with_invalid_id(nvme0, ncqa):
    def delete_sq(nvme0, qid):
        nvme0.send_cmd(0x00, cdw10 = qid).waitdone()

    # pass case
    cq = IOCQ(nvme0, 1, 10, PRP(4096))
    IOSQ(nvme0, 5, 10, PRP(4096), cqid=1)
    delete_sq(nvme0, 5)
    cq.delete()

    # sqid: 0
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        delete_sq(nvme0, 0)

    # sqid: 0xffff
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        delete_sq(nvme0, 0xffff)

    # sqid: larger than supported number of queue
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        delete_sq(nvme0, ncqa+1)

    # sqid: duplicated cqid
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        delete_sq(nvme0, ncqa+0xff)

    # sqid: not existed
    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        delete_sq(nvme0, 5)


def test_delete_cq_with_sq(nvme0, nsq=3, sqid=2):
    cq = IOCQ(nvme0, 2, 10, PRP(4096))
    sq_list = []
    for i in range(nsq):
        sq_list.append(IOSQ(nvme0, sqid+i, 10, PRP(4096), cqid=2))

    # delete cq before corresponding sq
    with pytest.warns(UserWarning, match="ERROR status: 01/0c"):
        cq.delete()

    for sq in sq_list:
        sq.delete()
    cq.delete()


def test_create_cq_with_invalid_queue_size(nvme0, mqes):
    if mqes == 0x10000:
        pytest.skip("no invalid queue size")
        
    with pytest.warns(UserWarning, match="ERROR status: 01/02"):
        IOCQ(nvme0, 1, 0xffff, PRP(4096))

    with pytest.warns(UserWarning, match="ERROR status: 01/02"):
        IOCQ(nvme0, 1, 1, PRP(4096))

    IOCQ(nvme0, 1, mqes-2, PRP(4096)).delete()
    IOCQ(nvme0, 1, mqes-1, PRP(4096)).delete()
    IOCQ(nvme0, 1, mqes, PRP(4096)).delete()

    with pytest.warns(UserWarning, match="ERROR status: 01/02"):
        IOCQ(nvme0, 1, mqes+1, PRP(4096))

    with pytest.warns(UserWarning, match="ERROR status: 01/02"):
        IOCQ(nvme0, 1, mqes+2, PRP(4096))

    with pytest.warns(UserWarning, match="ERROR status: 01/02"):
        IOCQ(nvme0, 1, mqes+0xff, PRP(4096))

    # empty queue
    with pytest.warns(UserWarning, match="ERROR status: 01/02"):
        IOCQ(nvme0, 1, 1, PRP(4096))
    with pytest.raises(AssertionError):
        IOCQ(nvme0, 1, 0, PRP(4096))

    IOCQ(nvme0, 1, 2, PRP(4096))


def test_create_sq_with_invalid_queue_size(nvme0, mqes):
    if mqes == 0x10000:
        pytest.skip("no invalid queue size")
        
    cq = IOCQ(nvme0, 2, 10, PRP(4096))

    with pytest.warns(UserWarning, match="ERROR status: 01/02"):
        IOSQ(nvme0, 1, 0xffff, PRP(4096), cqid=2)

    with pytest.warns(UserWarning, match="ERROR status: 01/02"):
        IOSQ(nvme0, 1, 1, PRP(4096), cqid=2)

    IOSQ(nvme0, 1, mqes-2, PRP(4096), cqid=2).delete()
    IOSQ(nvme0, 1, mqes-1, PRP(4096), cqid=2).delete()
    IOSQ(nvme0, 1, mqes, PRP(4096), cqid=2).delete()

    with pytest.warns(UserWarning, match="ERROR status: 01/02"):
        IOSQ(nvme0, 1, mqes+1, PRP(4096), cqid=2)

    with pytest.warns(UserWarning, match="ERROR status: 01/02"):
        IOSQ(nvme0, 1, mqes+2, PRP(4096), cqid=2)

    with pytest.warns(UserWarning, match="ERROR status: 01/02"):
        IOSQ(nvme0, 1, mqes+0xff, PRP(4096), cqid=2)

    # empty queue
    with pytest.warns(UserWarning, match="ERROR status: 01/02"):
        IOSQ(nvme0, 1, 1, PRP(4096), cqid=2)
    with pytest.raises(AssertionError):
        IOSQ(nvme0, 1, 0, PRP(4096), cqid=2)

    IOSQ(nvme0, 1, 2, PRP(4096), cqid=2).delete()
    cq.delete()


def test_create_sq_physically_contiguous(nvme0):
    if not nvme0.cap&0x10000:
        pytest.skip("pc is not required")

    cq = IOCQ(nvme0, 2, 10, PRP(4096))
    with pytest.warns(UserWarning, match="ERROR status: 00/02"):
        IOSQ(nvme0, 1, 2, PRP(4096), pc=False, cqid=2)
    IOSQ(nvme0, 1, 2, PRP(4096), pc=True, cqid=2).delete()
    cq.delete()


def test_create_sq_with_invalid_cqid(nvme0,mqes):
    cdw0=nvme0.getfeatures(0x07).waitdone()
    ncqr=(cdw0>>16)&0xffff

    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        IOSQ(nvme0, 1, mqes, PRP(4096), cqid=0)

    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        IOSQ(nvme0, 1, mqes, PRP(4096), cqid=0xffff)

    with pytest.warns(UserWarning, match="ERROR status: 01/00"):
        IOSQ(nvme0, 1, mqes, PRP(4096), cqid=ncqr)

    with pytest.warns(UserWarning, match="ERROR status: 01/00"):
        IOSQ(nvme0, 1, mqes, PRP(4096), cqid=ncqr+1)

    with pytest.warns(UserWarning, match="ERROR status: 01/01"):
        IOSQ(nvme0, 1, mqes, PRP(4096), cqid=ncqr+0xff)


def test_create_cq_invalid_interrupt_vector(nvme0, pcie):
    msicap_addr = pcie.cap_offset(0x05)
    msicap = pcie.register(msicap_addr, 4)
    logging.info("MSI Enable: %d" % ((msicap>>16)&0x1))
    logging.info("Multiple Message Capable: %d" % ((msicap>>17)&0x7))
    logging.info("Multiple Message Enable: %d" % ((msicap>>20)&0x7))
    if ((msicap>>20)&0x7) :
        invalid_iv = ((msicap>>17)&0x7) + 2
        logging.info("invalid_iv: %d" % invalid_iv)

    msixcap_addr = pcie.cap_offset(0x11)
    msixcap = pcie.register(msixcap_addr, 4)
    logging.info("Table Size: %d" % ((msixcap>>16)&0x3ff))
    logging.info("MSI-X Enable: %d" % ((msixcap>>31)&0x1))
    if ((msixcap>>31)&0x1) :
        invalid_iv = ((msixcap>>16)&0x3ff) + 2
        logging.info("invalid_iv: %d" % invalid_iv)

    with pytest.warns(UserWarning, match="ERROR status: 01/08"):
        IOCQ(nvme0, 1, 10, PRP(4096), iv=invalid_iv, ien=True)
    with pytest.warns(UserWarning, match="ERROR status: 01/08"):
        IOCQ(nvme0, 1, 10, PRP(4096), iv=2047, ien=True)


def test_create_cq_invalid_queue_address_offset(nvme0):
    queue = PRP(4096)
    queue.offset = 1
    with pytest.warns(UserWarning, match="ERROR status: 00/13"):
        IOCQ(nvme0, 1, 10, queue)


def test_create_sq_invalid_queue_address_offset(nvme0):
    cq = IOCQ(nvme0, 3, 10, PRP(4096))

    queue = PRP(4096)
    queue.offset = 1
    with pytest.warns(UserWarning, match="ERROR status: 00/13"):
        IOSQ(nvme0, 1, 10, queue, cqid=3)

    cq.delete()




# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem
from scripts.psd import IOCQ, IOSQ, PRP, PRPList, SQE, CQE


@pytest.fixture()
def cq(nvme0):
    ret = IOCQ(nvme0, 1, 10, PRP())
    yield ret
    ret.delete()

@pytest.fixture()
def sq(nvme0, cq):
    ret = IOSQ(nvme0, 1, 10, PRP(), cq.id)
    yield ret
    ret.delete()


def test_compare_lba_0(nvme0, nvme0n1, buf, qpair):
    if not nvme0n1.supports(5):
        pytest.skip("Compare command is not supported")

    ncap = nvme0n1.id_data(15, 8)
    nvme0n1.write(qpair, buf, 1, 1).waitdone()
    nvme0n1.write(qpair, buf, 0, 1).waitdone()
    nvme0n1.compare(qpair, buf, 0).waitdone()
    
    #Restore buf[0] value
    orig = buf[0]
    with pytest.warns(UserWarning, match="ERROR status: 02/85"):
        nvme0n1.compare(qpair, buf, ncap-1).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 02/85"):
        buf[0] += 1
        nvme0n1.compare(qpair, buf, 0).waitdone()
    buf[0] = orig
    with pytest.warns(UserWarning, match="ERROR status: 02/85"):
        nvme0n1.compare(qpair, buf, 0, 2).waitdone()

    with pytest.warns(UserWarning, match="ERROR status: 00/80"):
        nvme0n1.compare(qpair, buf, ncap).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/80"):
        nvme0n1.compare(qpair, buf, 0xffffffff).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/80"):
        nvme0n1.compare(qpair, buf, 0xffffffff00000000).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/80"):
        nvme0n1.compare(qpair, buf, 0x100000000).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/80"):
        nvme0n1.compare(qpair, buf, ncap-1, 2).waitdone()
    #ERROR status 00/02 or 00/80,it's depend on FW
    with pytest.warns(UserWarning, match="ERROR status: 00/(02|80)"):
        nvme0n1.compare(qpair, buf, ncap, 0x1000).waitdone()


@pytest.mark.parametrize("nsid", [0, 2, 3, 128, 255, 0xffff, 0xfffffffe, 0xffffffff])
def test_compare_invalid_nsid(nvme0, nvme0n1, nsid, cq, sq):
    # first cmd compare, invalid namespace
    cmd = SQE(5, nsid)
    buf = PRP(512)
    cmd.prp1 = buf
    sq[0] = cmd
    sq.tail = 1
    time.sleep(0.1)
    status = (cq[0][3]>>17)&0x7ff
    assert status == 0x000b or status == 0x0002


def test_fused_operations(nvme0, nvme0n1, qpair, buf):
    # check if fused commands supported
    if nvme0.id_data(523, 522) == 0:
        pytest.skip("fused command is not supported")

    # not fused,compare and write as separate command
    nvme0n1.write(qpair, buf, 8).waitdone()
    nvme0n1.compare(qpair, buf, 8).waitdone()
    # check if there is any Media and Data Integrity Errors
    with pytest.warns(UserWarning, match="ERROR status: 02/85"):
        logging.info("Compare failure!")

    # fused with correct order, compare 1st as the 1st cmd
    nvme0n1.send_cmd(5|(1<<8), qpair, buf, 1, 8, 0, 0)
    nvme0n1.send_cmd(1|(1<<9), qpair, buf, 1, 8, 0, 0)
    qpair.waitdone(2)

    # atomic: first cmd should be timeout
    with pytest.warns(UserWarning, match="ERROR status: 00/0a"):
        # wrong order: send write cmd 1st as 1st cmd, should abort
        nvme0n1.send_cmd(1|(1<<8), qpair, buf, 1, 8, 0, 0).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/0a"):
        # wrong order: send compare cmd 2nd as 2nd cmd, should abort
        nvme0n1.send_cmd(5|(1<<9), qpair, buf, 1, 8, 0, 0).waitdone()


def test_fused_cmd_not_supported(nvme0, nvme0n1, qpair, buf):
    # check if fused commands supported
    if nvme0.id_data(523, 522) == 0:
        logging.info("fused command is not supported")

        # write to init buffer before compare
        nvme0n1.write(qpair, buf, 8).waitdone()

        # if fuse not supported, fuse command should abort with invalid field
        logging.info("fused command is not supported, abort!")
        with pytest.warns(UserWarning, match="ERROR status: 00/02"):
            nvme0n1.send_cmd(5|(1<<8), qpair, buf, 1, 8, 0, 0)
            nvme0n1.send_cmd(1|(1<<9), qpair, buf, 1, 8, 0, 0)
            qpair.waitdone(2)
